{% extends "!layout.html" %}

{%- block extrahead %}
{{ super() }}
<script type="text/javascript">
  var _gaq = _gaq || [];
  _gaq.push(['_setAccount', 'UA-18486793-1']);
  _gaq.push(['_setDomainName', '.fabfile.org']);
  _gaq.push(['_trackPageview']);

  (function() {
    var ga = document.createElement('script'); ga.type = 'text/javascript'; ga.async = true;
    ga.src = ('https:' == document.location.protocol ? 'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
    var s = document.getElementsByTagName('script')[0]; s.parentNode.insertBefore(ga, s);
  })();
</script>

{% endblock %}


{% block sidebarlogo %}
{{ super() }}
{% if fabric_tags %}
<h3>Project Versions</h3>
<ul id="sidebar_versions">
    {% for tag in fabric_tags %}
    {% if loop.index <= 10 %}
    <li><a href="/en/{{ tag }}/">{{ tag }}</a></li>
    {% endif %}
    {% endfor %}
</ul>
{% endif %}
{% endblock %}
